/* Authors: Prodromos Gerakios and Nikolaos Papaspyrou
 *
 * All rights reserved.
 */
//#define NDEBUG  //remove assert code
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sched.h>
#include <sys/syscall.h>
// DO NOT REMOVE __RUNTIME__
// MUST BE DEFINED BEFORE runtime.h
#define  ___RUNTIME__
#include "runtime.h"

#ifdef USE_BOEHM
#include "libatomic_ops/src/atomic_ops.h"
#endif

#define MAX_FUTURE_LOCKSET 256
#define MAX_CACHE          4

void futex_wake (int *addr, int count);
void futex_wait (int *addr, int val);
void simple_futex_wait(int *addr, int val);

#define cpu_relax() __asm__ __volatile__ ("rep; nop" ::: "memory")

pid_t gettid ()
{
  return syscall(SYS_gettid);
}

#ifdef DEBUG
int detail_level = -1;
__thread char debug_buff[4096];
#define dprintf(detail, ...)                                            \
  do {                                                                  \
    if (detail <= detail_level) {                                       \
      char *dbg = debug_buff;                                           \
      dbg += sprintf(dbg, "[%s:%d] ", __FILE__, __LINE__);              \
      dbg += sprintf(dbg, "[pid=%u] ", (unsigned) gthread_pid);         \
      sprintf(dbg, __VA_ARGS__);                                        \
      puts(debug_buff);                                                 \
      fflush(stdout);                                                   \
    }                                                                   \
  } while(0)

#define ddprintf(detail, ...)                   \
  do {                                          \
    if (detail <= detail_level) {               \
      char *dbg = debug_buff;                   \
      sprintf(dbg, __VA_ARGS__);                \
      puts(debug_buff);                         \
      fflush(stdout);                           \
    }                                           \
  } while(0)
#else
#define dprintf(...)
#define ddprintf(...)
#endif

#define implies(a,b) (!(a) || (b))

typedef struct future_lockset {
  mypthread_mutex_t *mutex;
} future_lockset_t;

__thread pid_t gthread_pid = -1;
__thread stack_node_t *stack = 0;
__thread future_lockset_t future_lockset[MAX_FUTURE_LOCKSET];

__thread future_lockset_t *gfl_iter = 0;
__thread mypthread_mutex_t *glock = 0;
__thread int glocks_found = 0;
__thread stack_node_t * gstack_iter = 0;
__thread int  gignore_first = 0;

int compute_lockset_frame (node_t *iter_node,
                           ulong offset,
                           int locks_found,
                           node_t *pop_until){
  node_t *iter_node_tmp;
  node_t *iter_path,*iter_path_end;
  int path_locks,ret_locks;
  node_t n;

  assert(locks_found);
  assert(iter_node);
  assert(glock);
  assert((void *) gstack_iter > (void *) 100);

  while( iter_node && locks_found ){
   n = *iter_node;
   dprintf(2,"VISITING %s (%p) ",iter_node->node_name,iter_node);
   assert(iter_node->node_name);
   switch(n.meta_info.tag){
     case SIMPLE:
      assert(offset == 0);
      // find handle
      mypthread_mutex_t *h;
      switch(n.meta_info.mem){
       case STACK:
        h = ((mypthread_mutex_t *)(gstack_iter+1)) + n.kind.stack_offset;
        ddprintf(2," stackframe: (%p) + %d = %p. ",gstack_iter,
                 n.kind.stack_offset,h);
        break;
      case GLOBAL:
         h = n.kind.handle;
         ddprintf(2," global. ");
//#ifdef DEBUG
//			if(h && !h->name) h->name = iter_node->node_name;
//#endif
         break;
      case HEAP:
         h = n.kind.handle;
         ddprintf(2," heap. ");
         break;
      default:
         assert(0); //impossible!
         break;
      }
      int type = (int) n.meta_info.type;
      if( h == glock ){
        if( gignore_first ) gignore_first = 0;
        else  locks_found += type;
      }
      // add to lockset
      else if ( type == LOCK && h ) {
       if( gfl_iter >= future_lockset + MAX_FUTURE_LOCKSET)
       exit(-1);
       (gfl_iter++)->mutex = h;
         ddprintf(2," LOCKSET ");
		  h->name = (char *)iter_node->node_name;
         //dprintf("name : %s\n",h->name );
      }
		ddprintf(2,"\n");
      if( n.meta_info.parent_offset+1 < n.meta_info.size){
       iter_node++; //next elt in sequence
       //dprintf("INCREMENTED iter_node = %p\n",iter_node);
       continue;
      }
      break;
     ///  ///    //       ///     ///       ///    //
     case PATH:
      for
      (ret_locks=-1,
       path_locks = locks_found,
       iter_path = n.kind.com + offset,
       iter_path_end = iter_path + n.meta_info.size;
       //condition
        locks_found &&
       iter_path < iter_path_end ;
       // repeat instr
       path_locks = locks_found,//reset path_locks
       iter_path++ )
       {
        dprintf(3,"ENTERING PATH\n\n");
        path_locks = compute_lockset_frame
                     (iter_path,0,
                      path_locks,
                      iter_node // pop until this node
                     );
        dprintf(3,"EXITING PATH\n\n");
         if( ret_locks == -1 )
          ret_locks = path_locks;
         else
          assert(path_locks == ret_locks);
       }
      //assuming that all paths yield the same lock count
      locks_found = ret_locks;
      break;
     case SEQ:
      if( offset < n.meta_info.size ){
       dprintf(3,"LOOPING SEQ: offset=%ld size=%ld this-child=%ld\n\n",
               (long)offset,
               (long)n.meta_info.size,
               (long) (n.kind.com - iter_node));
       iter_node = n.kind.com + offset;
       offset = 0;
       continue;
      }
      break;
      // /// //
     case CALL:
       if( n.kind.com ){ // optimization: a null pointer is passed for functions with empty effects
        dprintf(3,"ENTERING CALL:\n\n");
        locks_found = compute_lockset_frame
                     (n.kind.com,0,locks_found,0);
        dprintf(3,"EXITING CALL\n\n");
       }
       break;
     default:
         assert(0); //impossible!
         break;
    } //

    dprintf(4,"starting with OLD node %s | POP UNTIL = %s\n",
             iter_node->node_name,
				 pop_until?pop_until->node_name:"next path");

    // go some levels up if necessary
	 int cond[3] = {0,0,0};
    do {
     if(!n.meta_info.has_parent)
      return locks_found;
     // parent node is n places back
     iter_node_tmp = iter_node - n.meta_info.parent;
     dprintf(4,"C1=%d C2=%d C3=%d\nThis->node = %s parent_offset=%d\n",
			  		cond[0],cond[1],cond[2],
              iter_node->node_name,(int)  n.meta_info.parent_offset);
     offset = n.meta_info.parent_offset; //this elt
     assert(offset <  iter_node_tmp->meta_info.size);
     offset++; // next elt
     dprintf(4,"next = %s next->size=%ld this->offset+1=%ld\t"
             "Skipping this->node : %s\n",
             iter_node_tmp->node_name ,
             (long) iter_node_tmp->meta_info.size,
             (long) offset,
             iter_node->node_name
            );

     iter_node = iter_node_tmp;
     n = *iter_node;

    } while( (
				 	(cond[0]=(offset >= n.meta_info.size))
					||
					(cond[1]=(n.meta_info.tag == PATH))
				 )
				 &&
				 (cond[2]=implies(pop_until,iter_node != pop_until))
             //||
             //(cond[1]=(!pop_until &&
             // n.meta_info.tag == PATH)))
             //||
             //(cond[2]=(pop_until &&
             // iter_node != pop_until)
             //)
           );
     dprintf(4,"C1=%d C2=%d C3=%d\nNEW iter_node = %s offset=%ld\n",
			  		cond[0],cond[1],cond[2],
             iter_node->node_name,(long) offset);
     if( iter_node == pop_until) break;
  }
  return locks_found;
}

#ifdef USE_BOEHM
#define __sync_fetch_and_add(a,b) AO_fetch_and_add_full(a,b)
#define __sync_fetch_and_sub(a,b) AO_fetch_and_add_full(a,-b)
#endif

void compute_lockset ()
{
  assert(glock);
  assert(glocks_found);
  assert((void *)stack > (void *)100);
  assert(stack->current);
  assert(stack->current->meta_info.type == LOCK);

  if(!glock->name){
   glock->name = (char *) stack->current->node_name;
  }

  dprintf(1,"lockset computation for: %s (%p)\n" ,
           glock->name, glock);

  // Iterate over the stack frames
  for(gstack_iter = stack;
      glocks_found && gstack_iter ;
      gstack_iter = gstack_iter->next){
    glocks_found =
        compute_lockset_frame(gstack_iter->current,
                              0,//gstack_iter->offset,
                              glocks_found,0);
  }
  dprintf(1,"lockset computation for %s (%p) done.\n" ,
	  glock->name, glock);
#ifdef DEBUG
  future_lockset_t * cu;
  dprintf(1,"LOCKSET: ");
  for( cu = future_lockset; cu < gfl_iter  ; cu++){
    dprintf(1,"%s ||", cu->mutex?cu->mutex->name:"(null)");
  }
  dprintf(1,"\n");
#endif
}

int mypthread_mutex_lock (mypthread_mutex_t *lock)
{
#ifdef DEBUG
  if (detail_level == -1) {
    char *env = getenv("DETAIL_LEVEL"), c;
    if (!env)
      detail_level = 0;
    else {
      c = *env;
      detail_level = (c < '1' || c > '9') ? 0 : c - '0';
    }
  }
#endif

  if (!lock) return 0;

  // store my own thread id for future use
  if (gthread_pid == -1) gthread_pid = gettid();

  // already locked by me: no need to check future lockset
  if (((pthread_mutex_t *)lock)->__data.__owner == gthread_pid) {
    dprintf(0, "LOCK ALREADY HELD for %u on %s\n",
            (unsigned) gthread_pid,
            lock ? lock->name : "(null)");
    return pthread_mutex_lock((pthread_mutex_t *) lock);
  }

  // compute future lockset
  glock = lock;               // DO NOT MOVE req by compute_lockset
  glocks_found = 1;           // DO NOT MOVE req by compute_lockset
  gfl_iter = future_lockset;  // end of future lockset
  gignore_first = 1;
  compute_lockset();

  future_lockset_t *cu;
  int lock_val, ret;
#ifdef __FUTEX__
  pthread_mutex_t *wt = NULL;
#endif

  // loop until successful
 RETRY:
  // check that all locks in future lockset are now available
  for (cu = future_lockset; cu < gfl_iter; cu++) {
    if (cu->mutex == NULL) continue;
    if (cu->mutex->mutex.__data.__owner == gthread_pid) {
      cu->mutex = NULL; // already owned by me, ignore from now on
      continue;
    }
    lock_val = __sync_fetch_and_add(&cu->mutex->mutex.__data.__lock, 0);
    if (lock_val != 0) {
#ifdef __FUTEX__
      if (wt != NULL) futex_wake(&wt->__data.__lock, 1);
      wt = (pthread_mutex_t *) cu->mutex;
#endif
      dprintf(0, "future lock %s not available (value %d)\n",
              cu->mutex ? ((mypthread_mutex_t *) cu->mutex)->name : "(null)",
              lock_val);
      goto YIELD;
    }
  }
  // lock tentatively (possibly suspending)
#ifdef __FUTEX__
  if (wt != NULL) { futex_wake(&wt->__data.__lock, 1); wt = NULL; }
#endif
  ret = pthread_mutex_lock((pthread_mutex_t *) lock);
  if (ret != 0) return ret;
  dprintf(0, "wanted lock %s acquired tentatively\n",
          lock ? ((mypthread_mutex_t *) lock)->name : "(null)");
  // check that versions did not change
  for (cu = future_lockset; cu < gfl_iter; cu++) {
    if (cu->mutex == NULL) continue;
    if (cu->mutex->mutex.__data.__owner == gthread_pid) continue;
    lock_val = __sync_fetch_and_add(&cu->mutex->mutex.__data.__lock, 0);
    if (lock_val != 0) {
      dprintf(0, "future lock %s not available (value %d)\n",
              cu->mutex ? ((mypthread_mutex_t *) cu->mutex)->name : "(null)",
              lock_val);
      pthread_mutex_unlock((pthread_mutex_t *) lock);
      dprintf(0, "wanted lock %s released\n",
              lock ? ((mypthread_mutex_t *) lock)->name : "(null)");
#ifdef __FUTEX__
      if (wt != NULL) futex_wake(&wt->__data.__lock, 1);
      wt = (pthread_mutex_t *) cu->mutex;
#endif
      goto YIELD;
    }
  }
  // successful
#ifdef __FUTEX__
  if (wt != NULL) { futex_wake(&wt->__data.__lock, 1); wt = NULL; }
#endif
  dprintf(0, "LOCK SUCCESSFUL on %s\n", lock ? lock->name : "(null)");
  return 0;
 YIELD:
#ifdef __FUTEX__
  if (lock_val != 2) {
    __sync_bool_compare_and_swap(&wt->__data.__lock, 1, 2);
    lock_val = 2;
  }
  dprintf(0, "FUTEX on %s value %d\n",
          wt ? ((mypthread_mutex_t *) wt)->name : "(null)", lock_val);
  futex_wait(&wt->__data.__lock, lock_val);
  dprintf(0, "FUTEX woken\n");
#else
#ifdef __YIELD__
  dprintf(0, "YIELDING\n");
  sched_yield();
#else
#ifdef __BUSY__
  cpu_relax();
#else
#error "Define exactly one of __FUTEX__, __YIELD__ or __BUSY__"
#endif
#endif
#endif
  goto RETRY;
}

int mypthread_mutex_unlock (mypthread_mutex_t *lock)
{
  int ret = pthread_mutex_unlock((pthread_mutex_t *) lock);
  dprintf(0, "UNLOCK SUCCESSFUL on %s\n", lock ? lock->name : "(null)");
  return ret;
}

int mypthread_cond_wait(pthread_cond_t *cond, mypthread_mutex_t *mutex)
{
  int ret = pthread_cond_wait(cond, (pthread_mutex_t *) mutex);
  mypthread_mutex_unlock(mutex);
  mypthread_mutex_lock(mutex);
  return ret;
}
